
# from Dashboards.Dashboards import Dashboards
# from Dashboards.Dashboards import Telemetry
# from Dashboards.Telemetry.Telemetry import Telemetry
# from Dashboards.InclusionExclusion.InclusionExclusion import InclusionExclusion




class ProductUsage:

    def get_base_query(self
                       , filters: list | None = None
                       , start_date: str = '1900-01-01'
                       , end_date: str = '9999-12-31'
                       ):

        import configparser

        config = configparser.ConfigParser()
        config.read('pip.ini')

        table = config['telemetry_product_usage'].get('table')
        a = config['telemetry_product_usage'].get('col_a')
        b = config['telemetry_product_usage'].get('col_b')
        c = config['telemetry_product_usage'].get('col_c')
        d = config['telemetry_product_usage'].get('col_d')
        e = config['telemetry_product_usage'].get('col_e')
        f = config['telemetry_product_usage'].get('col_f')
        g = config['telemetry_product_usage'].get('col_g')
        date_column = config['telemetry_product_usage'].get('date_column')
        filter_a = config['telemetry_product_usage'].get('filter_a')
        filter_b = config['telemetry_product_usage'].get('filter_b')
        filter_c = config['telemetry_product_usage'].get('filter_c')

        # Ensure filters is a list
        if not (isinstance(filters, list) or filters is None):
            raise TypeError("filters must be a list")

        if filters is None:
            filters = []
            filters_str = f'WHERE {date_column} BETWEEN {start_date} AND {end_date}'

        # Join column filters into comma-separated string, if any
        if len(filters) > 0:
            filters_str = 'AND '.join(filters)
            filters_str = f"WHERE {date_column} BETWEEN '{start_date}' AND '{end_date}' AND {filters_str}"
        else:
            filters_str = f"WHERE {date_column} BETWEEN '{start_date}' AND '{end_date}'"

        base_query = f"""
            SELECT
                DISTINCT 
                {a}
                , {b}
                , CASE 
                    WHEN {b} = '{filter_a}' THEN {c}
                    WHEN {b} = '{filter_b}' THEN {d}
                    WHEN {b} = '{filter_c}' THEN {e}
                    else null 
                END AS PRODUCT_NAME
                , COUNT(DISTINCT {f})
                , (SUM({g})/NULLIF(COUNT(DISTINCT {f}),0 )) AS DAILY_USAGE
            FROM 
                {table}
            
            {filters_str}
            
            GROUP BY
                {a}
                , {b}
                , PRODUCT_NAME
    """
        return base_query

    def product_viewed(self
                       , username: str
                       , warehouse: str
                       , database: str
                       , role: str
                       , filters: list | None = None
                       , order_by: list | None = None
                       , polars: bool = False
                       , start_date: str = '1900-01-01'
                       , end_date: str = '9999-12-31'
                       ):

        from NikeCA import Snowflake
        import polars as pl
        import configparser

        config = configparser.ConfigParser()
        config.read('pip.ini')

        b = config['telemetry_product_usage'].get('col_b')

        # Ensure filters is a list
        if not (isinstance(filters, list) or filters is None):
            raise TypeError("filters must be a list")

        # Ensure order_by is a list
        if not (isinstance(order_by, list) or order_by is None):
            raise TypeError("order_by must be a list")

        if order_by is None:
            order_by = config['telemetry_product_usage'].get('order_by')
        else:
            order_by = ', '.join(order_by)

        query = f"""
            /*
            The query calculates the daily usage of different data products on different platforms, groups the data by platform and product, and sorts the result in descending order based on the total usage
            */
            SELECT 
                {b}
                , PRODUCT_NAME
                , SUM(DAILY_USAGE) AS TOTAL_USAGE
            FROM (

            {self.get_base_query(start_date=start_date, end_date=end_date)}
            
            )
            
            GROUP BY 
                {b}
                , PRODUCT_NAME
                
            HAVING 
                TOTAL_USAGE > 0
                
            ORDER BY
                {order_by}
                
            ;
        """

        df = pl.DataFrame(Snowflake(username=username, warehouse=warehouse, database=database,
                                    role=role).snowflake_pull(query=query, polars=polars))

        print(query)

        return df

    def summary(self
                , username: str
                , warehouse: str
                , database: str
                , role: str
                , filters: list | None = None
                , order_by: list | None = None
                , polars: bool = False
                , start_date: str = '1900-01-01'
                , end_date: str = '9999-12-31'
                ):
        import configparser

        config = configparser.ConfigParser()
        config.read('pip.ini')
        b = config['telemetry_product_usage'].get('col_b')

        if order_by is None:
            order_by = config['telemetry_product_usage'].get('order_by')
        else:
            order_by = ', '.join(order_by)

        query = f"""
    SELECT 
        DISTINCT
        {b}
        , SUM(DAILY_USAGE) OVER (PARTITION BY {b}) AS TOTAL_USAGE
    FROM (
        {self.get_base_query(start_date=start_date, end_date=end_date, filters=filters)}   
    )
    ORDER BY
        {order_by}
"""

        from NikeCA import Snowflake
        import polars as pl

        df = pl.DataFrame(Snowflake(username=username, warehouse=warehouse, database=database,
                                    role=role).snowflake_pull(query=query, polars=polars))

        print(query)

        return df







class GitHub:

    def github_dependencies_loop(self, tables, headers, df, org, g, save_path):
        """
        Retrieve code files from GitHub search and store them in a dictionary, for each table in the given list of tables.

        Args:
            tables (list): A list of strings representing the names of the tables to be searched.
            headers (dict): A dictionary containing the GitHub authorization token.
            df (dict): A dictionary to store the search results.
            org (str): A string representing the organization to be searched.
            g: An instance of the GitHub class.
            save_path (str): A string representing the path to save the search results.

        Returns:
            dict: A dictionary containing the search results.

        Raises:
            Exception: If an error occurs while searching for the tables.

        Examples:
            >>> headers = {'Authorization': 'token abc123'}
            >>> tables = ['table1', 'table2']
            >>> org = 'my_org'
            >>> g = GitHub()
            >>> save_path = 'results.json'
            >>> df = {}
            >>> res = g.github_dependencies_loop(tables, headers, df, org, g, save_path)
            >>> type(res)
            <class 'dict'>
        """

        import requests
        import random
        import time
        import re
        import json

        n = 3
        chunks = [tables[i:i + n] for i in range(0, len(tables), n)]

        lucky = 0
        counter = 1
        for chunk in chunks:

            used = requests.get('https://api.github.com/rate_limit', headers=headers).json()

            while used['resources']['search']['remaining'] < 3:
                lucky = 0
                dice = random.choice([2, 3, 4, 6, 7, 9, 10, 11, 12])

                print(f"\nWant to play?\nLet's roll the dice: \n\tYou rolled a {dice}\n\nIn the Jungle you must wait "
                      f"until the dice roll 5 or 8")

                print('\nYou must wait ' + str(int(used['resources']['search']['reset']) - int(time.time()) + 3) +
                      ' seconds before rolling again.\n')

                time_now = int(time.time())

                if int(used['resources']['search']['reset']) - int(time_now) > 0:
                    time.sleep(int(used['resources']['search']['reset']) - (time.time()) + 3)

                used = requests.get('https://api.github.com/rate_limit', headers=headers).json()

            if lucky == 0:
                lucky = 5
                dice = random.choice([5, 8])

                print(f"\nLet's roll the dice: \n\tYou rolled a {dice}\nPlease proceed:\n")

            for table in chunk:
                if table in df['TABLES_SEARCHED']:
                    if df['TABLES_NOT_SEARCHED'] is not None:
                        df['TABLES_NOT_SEARCHED'] = df['TABLES_NOT_SEARCHED'].remove(table)
                    continue

                print(table)
                if org is not None:
                    search = g.search_code('org:' + org + ' ' + table)
                else:
                    search = g.search_code(table)

                print(f'Searching {org} {str(table)}:\t\t\t\ttable number: {str(counter)}')

                df[table] = {}

                for file in search:
                    result = requests.get(file.download_url)

                    result = result.text

                    indexes = [index.start() for index in re.finditer(table.upper(), result.upper())]

                    df[table][str(file.repository.name) + '/' + str(file.path)] = {}
                    df[table][str(file.repository.name) + '/' + str(file.path)]['Start'] = result[:1000]
                    df[table][str(file.repository.name) + '/' + str(file.path)]['char_index'] = {}

                    instance = 1
                    for index in indexes:
                        if index < 250:
                            index = 250
                        instance += 1

                        df[table][str(file.repository.name) + '/' + str(file.path)]['char_index'][str(index)] = \
                            result[index - 250: index + 250]

                time.sleep(2.001)

                counter += 1

                df['TABLES_SEARCHED'].append(table)

                if df['TABLES_NOT_SEARCHED'] is not None:
                    df['TABLES_NOT_SEARCHED'] = df['TABLES_NOT_SEARCHED'].remove(table)

            if save_path is not None:
                with open(save_path, "w+") as f:
                    json.dump(df, f, indent=4)

        return df

    def github_dependencies(self, tables: list | str = None, token: str = None,
                            save_path: str | None = None, org: str | None = None):

        """
        Searches for dependencies of specified tables on GitHub and returns a dictionary with the dependencies.

        Args:
            tables (list or str, optional): A list of table names or a single table name. Defaults to None.
            token (str, optional): GitHub API token. Defaults to None.
            save_path (str, optional): Path to save the results. Defaults to None.
            org (str, optional): GitHub organization name. Defaults to None.

        Returns:
            dict: A dictionary with the dependencies of the specified tables on GitHub.

        Raises:
            StopIteration: Raised when there are no more tables left to search for.
            Exception: Raised when a general exception occurs.

        Examples:

        # Search for dependencies of a single table
        >>> dependencies = github_dependencies(tables='table1', token='your_github_token')
        Searching for 1 tables in GitHub

        # Search for dependencies of multiple tables
        >>> dependencies = github_dependencies(tables=['table1', 'table2'], token='your_github_token')
        Searching for 2 tables in GitHub

        # Search for dependencies of multiple tables with an organization name
        >>> dependencies = github_dependencies(tables=['table1', 'table2'], token='your_github_token', org='your_org_name')
        Searching for 2 tables in your_org_name

        # Save results to a file
        >>> dependencies = github_dependencies(tables='table1', token='your_github_token', save_path='dependencies.json')
        Searching for 1 tables in GitHub
        """

        import pandas as pd
        import time
        import json
        import os
        from github import Github as g
        from warnings import simplefilter
        import random

        simplefilter(action="ignore", category=pd.errors.PerformanceWarning)

        if tables is str:
            tables = [tables]

        gt = g(token)
        headers = {'Authorization': 'token ' + token}

        tables = [*set(tables)]
        print('Searching for ' + str(len(tables)) + ' tables in github')

        if save_path is not None and os.path.exists(save_path) and os.path.getsize(save_path) > 0:
            try:
                with open(save_path, 'r') as f:
                    df = json.load(f)
            except Exception as e:
                df = {}
                print(e)
        else:
            df = {}

        if 'TABLES_SEARCHED' not in df.keys():
            df['TABLES_SEARCHED'] = []
        if 'TABLES_NOT_SEARCHED' not in df.keys() or df['TABLES_NOT_SEARCHED'] is None:
            df['TABLES_NOT_SEARCHED'] = tables
        else:
            df['TABLES_NOT_SEARCHED'].append(tables)

        while len(df['TABLES_SEARCHED']) < len(tables):
            try:
                d_dict = GitHub.github_dependencies_loop(self, tables=tables, headers=headers, df=df, org=org, g=gt,
                                                         save_path=save_path)

                if save_path is None:
                    return d_dict

            except StopIteration:
                with open(save_path, 'r') as f:
                    d = json.load(f)
                return d
            except Exception as e:

                dice = random.choice([2, 3, 4, 6, 7, 9, 10, 11, 12])

                print(f"\n\nWant to play?\nLet's roll the dice: \n\tYou rolled a {dice}\n\nIn the Jungle you must wait "
                      f"until the dice roll 5 or 8\n\n")
                print(e)

                print('\n\n\nA General exception has occurred.\n\nWe will be put in the penalty box for 5 minutes\n\n')

                time.sleep(60*5)

                dice = random.choice([5, 8])

                print(f"\nLet's roll the dice: \n\tYou rolled a {dice}\nPlease proceed:\n")

                print("\n\nWe are free!!\n\nIt's go time!!\n")
                continue
        with open(save_path, 'r') as f:
            d = json.load(f)
        return d

    def get_repo_list(self, git_username: str = None, pat: str | None = None, org_name: str | None = None,
                      repo_list_filename: str | None = None):
        import json
        import requests
        from requests.auth import HTTPBasicAuth
        # import links_from_header

        respget = lambda url: requests.get(url, auth=HTTPBasicAuth(git_username, pat))

        myorgname = org_name
        nexturl = f"https://api.github.com/orgs/{myorgname}/repos?per_page=100"

        while nexturl:
            print(nexturl)
            resp = respget(nexturl)
            linkheads = resp.headers.get('Link', None)
            if linkheads:
                linkheads_parsed = links_from_header.extract(linkheads)
                nexturl = linkheads_parsed.get('next', None)
            else:
                nexturl = None
            respcon = json.loads(resp.content)

            with open(repo_list_filename, 'w+') as fh:
                fh.writelines([f'{respconi["full_name"]}\n' for respconi in respcon])

    def get_clone(self, git_username: str = None, pat: str = None, filename: str = None, org_name: str = None,
                  repo_name: str = None, repo_list_filename: str = None, repo_directory_name: str | None = None):
        """
        This function clones a repository from GitHub using the username, password/Personal Acess Token (PAT), and
        either a filename or org_name and repo_name. The repository is saved in the savepath directory.

        Parameters
        ----------
        username : str
            The username of the GitHub account.
        pat : str
            The Personal Access Token (PAT) of the GitHub account.
        filename : str, optional
            The name of the file containing a list of orgs and repo names in the format of <org name>/<repo name> per
            line.
        org_name : str, optional
            The name of the organization hosting the repository.
        repo_name : str, optional
            The name of the repository.
        savepath : str, optional
            The path of the directory where the repository will be saved.

        Returns
        -------
        None
            The function does not return anything.

        Examples
        --------
        >>> git_clone('username', 'password', filename='repos.txt', savepath='/home/user/repos')
        Cloning repository from github.com/org_name/repo_name
        Repository cloned successfully!
        :param org_name:
        :param repo_name:
        :param pat:
        :param filename:
        :param git_username:
        :param repo_directory_name:
        :param repo_list_filename:
        """

        import os

        # Check if filename is provided
        if filename is not None:
            # Read the file
            try:
                with open(filename, 'r') as f:
                    repos = f.readlines()
            except FileNotFoundError:
                print('File not found.')
                return
            # Iterate through the list of repos
            for repo in repos:
                org_name, repo_name = repo.strip().split('/')
                # Create a new directory with the repo name
                if repo_list_filename is not None:
                    try:
                        os.mkdir(os.path.join(repo_directory_name, repo_name))
                    except FileExistsError:
                        print('Directory already exists.')
                        print(repo)
                        continue
                # Clone the repository
                try:
                    os.chdir(os.path.join(repo_directory_name, repo_name))
                    os.system(f'git clone https://{git_username}:{pat}@github.com/{org_name}/{repo_name}.git')
                    print(f'Cloning repository from github.com/{org_name}/{repo_name}')
                    print('Repository cloned successfully!')
                except Exception as e:
                    print(e)
                    return
        # If filename is not provided, use org_name and repo_name
        else:
            # Create a new directory with the repo name
            if repo_list_filename is not None:
                try:
                    os.mkdir(os.path.join(repo_list_filename, repo_name))
                except FileExistsError:
                    print('Directory already exists.')
                    pass
            # Clone the repository
            try:
                os.chdir(os.path.join(repo_directory_name, repo_name))
                os.system(f'git clone https://{git_username}:{pat}@github.com/{org_name}/{repo_name}.git')
                print(f'Cloning repository from github.com/{org_name}/{repo_name}')
                print('Repository cloned successfully!')
            except Exception as e:
                print(e)
                return




from NikeSF import Snowflake
from NikeQA import QA
from Dashboards.Dashboards import Dashboards
# from Dashboards.InclusionExclusion.InclusionExclusion import InclusionExclusion
# from Dashboards.Telemetry.Telemetry import Telemetry
# from Dashboards.Telemetry.ProductUsage import ProductUsage


class NikeCA(
    Snowflake
    , QA
    , Dashboards
    # , Telemetry
    # , InclusionExclusion
    # , ProductUsage
):
    pass




from .Telemetry.Telemetry import Telemetry
from .InclusionExclusion.InclusionExclusion import InclusionExclusion


class Dashboards(Telemetry, InclusionExclusion):

    Telemetry = Telemetry()
    Telemetry.ProductUsage = Telemetry.ProductUsage()
    InclusionExclusion = InclusionExclusion()

    def imp_summary_dashboard(self
                              , username: str
                              , warehouse: str
                              , database: str
                              , role: str
                              , date_min: str = '1900-01-01'
                              , date_max: str = '9999-12-31'
                              , column_filters=None
                              , polars: bool = False):

        """Create the data to match IMP Summary Dashboard.

        Args:
            username (str): The username for the Snowflake connection.
            warehouse (str): The warehouse for the Snowflake connection.
            database (str): The database for the Snowflake connection.
            role (str): The role for the Snowflake connection.
            date_min (str): The minimum date for the query (default is '1900-01-01').
            date_max (str): The maximum date for the query (default is '9999-12-31').
            column_filters (list): A list of column filters for the query.
            polars (bool): Whether to return the query results as a Polars DataFrame (default is False, which returns
                a Pandas DataFrame).

        Returns:
            A Pandas or Polars DataFrame containing the query results.

        Raises:
            TypeError: If column_filters is not a list.
        """

        from NikeCA import Snowflake

        if column_filters is None:
            column_filters = []
        import configparser

        config = configparser.ConfigParser()
        config.read('pip.ini')


        date_begin = config['imp_summary'].get('begin_date')
        date_end = config['imp_summary'].get('date_end')
        date = config['imp_summary'].get('date')
        table = config['imp_summary'].get('table')
        column = config['imp_summary'].get('column')
        column_filter = config['imp_summary'].get('column_filter')
        column_flag = config['imp_summary'].get('column_flag')
        column_flag_answer = config['imp_summary'].get('column_flag_answer')
        aggs = config['imp_summary'].get('aggs')

        # Ensure column_filters is a list
        if not isinstance(column_filters, list):
            raise TypeError("column_filters must be a list")

        # Join column filters into comma-separated string, if any
        if len(column_filters) > 0:
            column_filters_str = ', '.join(column_filters)
            group_column_filters_str = f'''GROUP BY 
                {column_filters_str}'''
            column_filters_str += ', '
        else:
            column_filters_str = column + ', '
            group_column_filters_str = f'GROUP BY {column}'

        query_dashboard = f"""
        SELECT
            {column_filters_str}
            MIN({date_begin}) AS MIN_{date_begin}
            , MAX({date_end}) AS MAX_{date_end}
            , {aggs}
        FROM 
            {table}
        WHERE
            {date} BETWEEN '{date_min}' AND '{date_max}'
            AND {column} = '{column_filter}'
        {group_column_filters_str} 
        UNION
    
        SELECT
            {column_filters_str}
            MIN({date_begin}) AS MIN_{date_begin}
            , MAX({date_end}) AS MAX_{date_end}
            , {aggs}
        FROM 
            {table}
        WHERE
            {date} BETWEEN '{date_min}' AND '{date_max}'
            AND {column} <> '{column_filter}'
            AND {column_flag} = '{column_flag_answer}'
        {group_column_filters_str}
        ;
        """
        print(query_dashboard)

        return Snowflake(username=username, warehouse=warehouse, database=database,
                         role=role).snowflake_pull(query=query_dashboard, polars=polars)








class SnowflakeDependencies:

    def snowflake_dependencies(self,
                               tables: str | list,
                               username: str,
                               warehouse: str,
                               role: str,
                               database: str | None = None,
                               schema: str | list | None = None,
                               save_path: str | None = None):

        """
        Searches the snowflake database and finds instances where the table is referenced and where the reference is not
            in the actual creation of the table itself

        :param save_path: (str | None) The path to save the output.
        :param tables: (str | list): The list of tables or views to fetch DDL for.
        :param username: username for Snowflake
        :param warehouse: warehouse for Snowflake
        :param role: role for Snowflake
        :param database: database to search in must provide
        :param schema: filling this in can really help to speed up the query
        :return: pandas Dataframe

        This function is used to fetch the DDL of tables and views from Snowflake.

        Parameters:
            tables (str | list): The list of tables or views to fetch DDL for.
            username (str): The username of the Snowflake user.
            warehouse (str): The warehouse to use for the connection.
            role (str): The role of the Snowflake user.
            database (str | None): The database to use for the connection.
            schema (str | list | None): The schema to use for the connection.
            save_path (str | None): The path to save the output.

        Returns:
            dict: A dictionary containing the DDL of the tables and views.
        """

        # snowflake connection packages:
        import pandas as pd
        import polars as pl
        import snowflake.connector
        import time
        import json
        import os

        if type(tables) == str:
            tables = [tables]

        print('opening snowflake connection...')

        cnn = snowflake.connector.connect(
            user=username,
            account='nike',
            authenticator='externalbrowser',
            role=role,
            warehouse=warehouse,
        )
        cs = cnn.cursor()
        process_complete = 0
        process_pass = 0
        counter = 0

        # fetch schema and table names
        query = 'SELECT * FROM ' + database + '.INFORMATION_SCHEMA.TABLES'
        if schema is not None:
            if type(schema) == str:
                schema = [schema]
            query = query + " WHERE TABLE_SCHEMA IN ('" + schema[0] + "'"
            if len(schema) > 1:
                for i in schema[1:]:
                    query = query + ", '" + i + "'"
            query = query + ')'

        cs.execute(query)
        df_tables = cs.fetch_pandas_all()

        query_ddl = {}

        for k, i in df_tables.iterrows():
            if i['TABLE_TYPE'] == 'VIEW':
                query_ddl[i['TABLE_CATALOG'] + '.' + i['TABLE_SCHEMA'] + '.' + i['TABLE_NAME']] = \
                    "SELECT GET_DDL('" + i['TABLE_TYPE'] + "', '" + i['TABLE_CATALOG'] + '.' + i[
                        'TABLE_SCHEMA'] + '.' \
                    + i['TABLE_NAME'] + "')"
            elif i['TABLE_TYPE'] == 'BASE TABLE':
                query_ddl[i['TABLE_CATALOG'] + '.' + i['TABLE_SCHEMA'] + '.' + i['TABLE_NAME']] = \
                    "SELECT GET_DDL('TABLE', '" + i['TABLE_CATALOG'] + '.' + i['TABLE_SCHEMA'] + '.' + \
                    i['TABLE_NAME'] + "')"

        df = pd.DataFrame([query_ddl]).T

        df_index = df.index

        df_return = pd.DataFrame(index=df_index)
        df_return['sfqid'] = ''

        queries = len(df)
        print('Pulling ' + str(queries) + ' queries')

        query_list = []
        db_list = []
        complete = []

        for item in range(queries):
            query_list.append(item)
            db_list.append(item)
            complete.append(0)

        for k, v in df.iterrows():
            sql = v[0]
            cs.execute_async(sql)
            query_list[counter] = cs.sfqid
            df_return['sfqid'][k] = cs.sfqid
            counter += 1
        counter = 1
        if save_path is not None:
            if os.path.isfile(save_path):
                with open(save_path) as f:
                    d = json.load(f)
            else:
                d = {}
        else:
            d = {}
        while process_complete == 0:
            item = -1
            process_pass += 1
            if sum(complete) == queries or process_pass == 1000:
                process_complete = 1
            for result in query_list:
                item += 1
                if complete[item] == 0:
                    print('Running ' + df_return[df_return['sfqid'] == result].index[0])
                    try:
                        status = cnn.get_query_status_throw_if_error(result)

                    except snowflake.connector.errors.ProgrammingError:
                        print(f"""Could not retrieve:
                            {df_return[df_return['sfqid'] == result].index[0]}
                            
                            because of snowflake.connector.errors.ProgrammingError""")
                        complete[item] = 1
                        continue

                    except TypeError:
                        print(f"""Could not retrieve:
                            {df_return[df_return['sfqid'] == result].index[0]}
                        
                        because TypeError: NoneType""")
                        complete[item] = 1
                        continue

                    print('the status for ' + df_return[df_return['sfqid'] == result].index[0] + ' is ' +
                          str(status) + ' ' + str(counter))
                    if str(status) == 'QueryStatus.SUCCESS':
                        complete[item] = 1
                        cs.get_results_from_sfqid(result)
                        data = cs.fetch_pandas_all()
                        for table in tables:
                            try:
                                if table.upper() + ' ' in data.iloc[0, 0][data.iloc[0, 0].upper().index(
                                        table.upper()):data.iloc[0, 0].upper().index(table.upper()) + len(table) + 1]\
                                        and table.upper() not in data.iloc[0, 0][:data.iloc[0, 0].index('(')]:
                                    d[table] = {}
                                    d[table][df_return[df_return['sfqid'] == result].index[0]] = data.to_dict('index')
                            except ValueError:
                                continue
                    else:
                        time.sleep(.1)
                counter += 1
        cnn.close()

        if save_path is not None:
            with open(save_path, "w+") as f:
                json.dump(d, f, indent=4)

        print('process complete')

        return d

    
